@@ -1,5 +1,5 @@
 # coding: utf8
-# Copyright 2019 Lenovo
+# Copyright 2021 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,7 +21,6 @@ for redfish compliant endpoints
 import base64
 from datetime import datetime
 from datetime import timedelta
-from fnmatch import fnmatch
 import json
 import os
 import re
@@ -51,22 +50,6 @@ powerstates = {
     'boot': None,
 }
 
-boot_devices_write = {
-    'net': 'Pxe',
-    'network': 'Pxe',
-    'pxe': 'Pxe',
-    'hd': 'Hdd',
-    'usb': 'Usb',
-    'cd': 'Cd',
-    'cdrom': 'Cd',
-    'optical': 'Cd',
-    'dvd': 'Cd',
-    'floppy': 'Floppy',
-    'default': 'None',
-    'setup': 'BiosSetup',
-    'bios': 'BiosSetup',
-    'f1': 'BiosSetup',
-}
 
 boot_devices_read = {
     'BiosSetup': 'setup',
@@ -159,80 +142,16 @@ class SensorReading(object):
         self.imprecision = None
         self.units = units
         self.unavailable = unavailable
-
-
-class AttrDependencyHandler(object):
-    def __init__(self, dependencies, currsettings, pendingsettings):
-        self.dependencymap = {}
-        for dep in dependencies.get('Dependencies', [{}]):
-            if 'Dependency' not in dep:
-                continue
-            if dep['Type'] != 'Map':
-                continue
-            if dep['DependencyFor'] in self.dependencymap:
-                self.dependencymap[
-                    dep['DependencyFor']].append(dep['Dependency'])
-            else:
-                self.dependencymap[
-                    dep['DependencyFor']] = [dep['Dependency']]
-        self.curr = currsettings
-        self.pend = pendingsettings
-        self.reg = dependencies['Attributes']
-
-    def get_overrides(self, setting):
-        overrides = {}
-        blameattrs = []
-        if setting not in self.dependencymap:
-            return {}, []
-        for depinfo in self.dependencymap[setting]:
-            lastoper = None
-            lastcond = None
-            for mapfrom in depinfo.get('MapFrom', []):
-                if lastcond is not None and not lastoper:
-                    break  # MapTerm required to make sense of this, give up
-                currattr = mapfrom['MapFromAttribute']
-                blameattrs.append(currattr)
-                currprop = mapfrom['MapFromProperty']
-                if currprop == 'CurrentValue':
-                    if currattr in self.pend:
-                        currval = self.pend[currattr]
-                    else:
-                        currval = self.curr[currattr]
-                else:
-                    currval = self.reg[currattr][currprop]
-                lastcond = self.process(currval, mapfrom, lastcond, lastoper)
-                lastoper = mapfrom.get('MapTerms', None)
-            if lastcond:
-                if setting not in overrides:
-                    overrides[setting] = {}
-                if depinfo['MapToAttribute'] not in overrides[setting]:
-                    overrides[depinfo['MapToAttribute']] = {}
-                overrides[depinfo['MapToAttribute']][
-                    depinfo['MapToProperty']] = depinfo['MapToValue']
-        return overrides, blameattrs
-
-    def process(self, currval, mapfrom, lastcond, lastoper):
-        newcond = None
-        mfc = mapfrom['MapFromCondition']
-        if mfc == 'EQU':
-            newcond = currval == mapfrom['MapFromValue']
-        if mfc == 'NEQ':
-            newcond = currval != mapfrom['MapFromValue']
-        if mfc == 'GEQ':
-            newcond = float(currval) >= float(mapfrom['MapFromValue'])
-        if mfc == 'GTR':
-            newcond = float(currval) > float(mapfrom['MapFromValue'])
-        if mfc == 'LEQ':
-            newcond = float(currval) <= float(mapfrom['MapFromValue'])
-        if mfc == 'LSS':
-            newcond = float(currval) < float(mapfrom['MapFromValue'])
-        if lastcond is not None:
-            if lastoper == 'AND':
-                return lastcond and newcond
-            elif lastoper == 'OR':
-                return lastcond or newcond
-            return None
-        return newcond
+    
+    def __repr__(self):
+        return repr({
+            'value': self.value,
+            'state_ids': self.state_ids,
+            'units': self.units,
+            'imprecision': self.imprecision,
+            'name': self.name,
+            'unavailable': self.unavailable,
+        })
 
 
 class Command(object):
@@ -254,6 +173,15 @@ class Command(object):
         self._varfwinventory = None
         self._oem = None
         self._gpool = pool
+        self._bmcv4ip = None
+        self._bmcv6ip = None
+        for addrinf in socket.getaddrinfo(bmc, 0, 0, socket.SOCK_STREAM):
+            if addrinf[0] == socket.AF_INET:
+                self._bmcv4ip = socket.inet_pton(addrinf[0], addrinf[-1][0])
+            elif addrinf[0] == socket.AF_INET6:
+                theip = addrinf[-1][0]
+                theip = theip.split('%')[0]
+                self._bmcv6ip = socket.inet_pton(addrinf[0], theip)
         self.wc.set_header('Accept', 'application/json')
         self.wc.set_header('User-Agent', 'pyghmi')
         self.wc.set_header('Accept-Encoding', 'gzip')
@@ -276,13 +204,15 @@ class Command(object):
         systems = members['Members']
         if sysurl:
             for system in systems:
-                if system['@odata.id'] == sysurl:
-                    self.sysurl = sysurl
+                if system['@odata.id'] == sysurl or system['@odata.id'].split('/')[-1] == sysurl:
+                    self.sysurl = system['@odata.id']
                     break
             else:
                 raise exc.PyghmiException(
                     'Specified sysurl not found: {0}'.format(sysurl))
         else:
+            if len(systems) != 1:
+                systems = [x for x in systems if 'DPU' not in x['@odata.id']]
             if len(systems) != 1:
                 raise exc.PyghmiException(
                     'Multi system manager, sysurl is required parameter')
@@ -436,6 +366,8 @@ class Command(object):
         return True
 
     def set_user_access(self, uid, privilege_level='ReadOnly'):
+        if privilege_level.startswith('custom.'):
+            privilege_level = privilege_level.replace('custom.', '')
         accinfo = self._account_url_info_by_id(uid)
         if not accinfo:
             raise Exception("Unable to find indicated uid")
@@ -457,6 +389,8 @@ class Command(object):
         accinfo = self._account_url_info_by_id(uid)
         if not accinfo:
             raise Exception("Unable to find indicated uid")
+        if privilege_level.startswith('custom.'):
+            privilege_level = privilege_level.replace('custom.', '')
         for role in self._validroles:
             if role.lower() == privilege_level.lower():
                 privilege_level = role
@@ -471,13 +405,7 @@ class Command(object):
         return True
 
     def user_delete(self, uid):
-        # Redfish doesn't do so well with Deleting users either...
-        # Blanking the username seems to be the convention
-        # First, set a bogus password in case the implementation does honor
-        # blank user, at least render such an account harmless
-        self.set_user_password(uid, base64.b64encode(os.urandom(15)))
-        self.set_user_name(uid, '')
-        return True
+        self.oem.user_delete(uid)
 
     def set_user_name(self, uid, name):
         """Set user name
@@ -591,6 +519,8 @@ class Command(object):
     def _do_web_request(self, url, payload=None, method=None, cache=True,
                         etag=None):
         res = None
+        if cache is True:
+            cache = 30
         if cache and payload is None and method is None:
             res = self._get_cache(url, cache)
         if res:
@@ -676,31 +606,7 @@ class Command(object):
         :raises: PyghmiException on an error.
         :returns: dict or True -- If callback is not provided, the response
         """
-        reqbootdev = bootdev
-        if (bootdev not in boot_devices_write
-                and bootdev not in boot_devices_read):
-            raise exc.InvalidParameterValue('Unsupported device %s'
-                                            % repr(bootdev))
-        bootdev = boot_devices_write.get(bootdev, bootdev)
-        if bootdev == 'None':
-            payload = {'Boot': {'BootSourceOverrideEnabled': 'Disabled'}}
-        else:
-            payload = {'Boot': {
-                'BootSourceOverrideEnabled': 'Continuous' if persist
-                                             else 'Once',
-                'BootSourceOverrideTarget': bootdev,
-            }}
-            if uefiboot is not None:
-                uefiboot = 'UEFI' if uefiboot else 'Legacy'
-                payload['BootSourceOverrideMode'] = uefiboot
-                try:
-                    self._do_web_request(self.sysurl, payload, method='PATCH')
-                    return {'bootdev': reqbootdev}
-                except Exception:
-                    del payload['BootSourceOverrideMode']
-        thetag = self.sysinfo.get('@odata.etag', None)
-        self._do_web_request(self.sysurl, payload, method='PATCH', etag=thetag)
-        return {'bootdev': reqbootdev}
+        return self.oem.set_bootdev(bootdev, persist, uefiboot, self)
 
     @property
     def _biosurl(self):
@@ -758,6 +664,14 @@ class Command(object):
         for subchassis in chassisinfo.get('Links', {}).get('Contains', []):
             self._mapchassissensors(subchassis)
 
+    def _get_thermals(self, chassis):
+        chassisurl = chassis['@odata.id']
+        chassisinfo = self._do_web_request(chassisurl)
+        thermurl = chassisinfo.get('Thermal', {}).get('@odata.id', '')
+        if thermurl:
+            therminf = self._do_web_request(thermurl, cache=1)
+            return therminf.get('Temperatures', [])
+
     @property
     def _bmcurl(self):
         if not self._varbmcurl:
@@ -807,6 +721,16 @@ class Command(object):
             if not nicinfo.get('InterfaceEnabled', True):
                 # skip disabled interfaces
                 continue
+            for addrs in nicinfo.get('IPv4Addresses', []):
+                v4addr = socket.inet_pton(
+                    socket.AF_INET, addrs.get('Address', '0.0.0.0'))
+                if self._bmcv4ip == v4addr:
+                    return curl
+            for addrs in nicinfo.get('IPv6Addresses', []):
+                v6addr = socket.inet_pton(
+                    socket.AF_INET6, addrs.get('Address', '::'))
+                if self._bmcv6ip == v6addr:
+                    return curl
             foundnics += 1
             lastnicurl = curl
         if name is None and foundnics != 1:
@@ -863,106 +787,7 @@ class Command(object):
         return {'identifystate': self._idstatemap[ledstate]}
 
     def get_health(self, verbose=True):
-        health = self.sysinfo.get('Status', {})
-        health = health.get('HealthRollup', health.get('Health', 'Unknown'))
-        warnunknown = health == 'Unknown'
-        health = _healthmap[health]
-        summary = {'badreadings': [], 'health': health}
-        if health > 0 and verbose:
-            # now have to manually peruse all psus, fans, processors, ram,
-            # storage
-            procsumstatus = self.sysinfo.get('ProcessorSummary', {}).get(
-                'Status', {})
-            procsumstatus = procsumstatus.get('HealthRollup',
-                                              procsumstatus.get('Health',
-                                                                None))
-            if procsumstatus != 'OK':
-                procfound = False
-                procurl = self.sysinfo.get('Processors', {}).get('@odata.id',
-                                                                 None)
-                if procurl:
-                    for cpu in self._do_web_request(procurl).get(
-                            'Members', []):
-                        cinfo = self._do_web_request(cpu['@odata.id'])
-                        if cinfo.get('Status', {}).get(
-                                'State', None) == 'Absent':
-                            continue
-                        if cinfo.get('Status', {}).get(
-                                'Health', None) not in ('OK', None):
-                            procfound = True
-                            summary['badreadings'].append(SensorReading(cinfo))
-                if not procfound:
-                    procinfo = self.sysinfo['ProcessorSummary']
-                    procinfo['Name'] = 'Processors'
-                    summary['badreadings'].append(SensorReading(procinfo))
-            memsumstatus = self.sysinfo.get(
-                'MemorySummary', {}).get('Status', {})
-            memsumstatus = memsumstatus.get('HealthRollup',
-                                            memsumstatus.get('Health', None))
-            if memsumstatus != 'OK':
-                dimmfound = False
-                for mem in self._do_web_request(
-                        self.sysinfo['Memory']['@odata.id'])['Members']:
-                    dimminfo = self._do_web_request(mem['@odata.id'])
-                    if dimminfo.get('Status', {}).get(
-                            'State', None) == 'Absent':
-                        continue
-                    if dimminfo.get('Status', {}).get(
-                            'Health', None) not in ('OK', None):
-                        summary['badreadings'].append(SensorReading(dimminfo))
-                        dimmfound = True
-                if not dimmfound:
-                    meminfo = self.sysinfo['MemorySummary']
-                    meminfo['Name'] = 'Memory'
-                    summary['badreadings'].append(SensorReading(meminfo))
-            for adapter in self.sysinfo['PCIeDevices']:
-                adpinfo = self._do_web_request(adapter['@odata.id'])
-                if adpinfo['Status']['Health'] not in ('OK', None):
-                    summary['badreadings'].append(SensorReading(adpinfo))
-            for fun in self.sysinfo['PCIeFunctions']:
-                funinfo = self._do_web_request(fun['@odata.id'])
-                if funinfo['Status']['Health'] not in ('OK', None):
-                    summary['badreadings'].append(SensorReading(funinfo))
-        if warnunknown and not summary['badreadings']:
-            unkinf = SensorReading({'Name': 'BMC',
-                                    'Status': {'Health': 'Unknown'}})
-            unkinf.states = ['System does not provide health information']
-            summary['badreadings'].append(unkinf)
-        return summary
-
-    def _get_biosreg(self, url):
-        addon = {}
-        valtodisplay = {}
-        displaytoval = {}
-        reg = self._do_web_request(url)
-        reg = reg['RegistryEntries']
-        for attr in reg['Attributes']:
-            vals = attr.get('Value', [])
-            if vals:
-                valtodisplay[attr['AttributeName']] = {}
-                displaytoval[attr['AttributeName']] = {}
-                for val in vals:
-                    valtodisplay[
-                        attr['AttributeName']][val['ValueName']] = val[
-                            'ValueDisplayName']
-                    displaytoval[
-                        attr['AttributeName']][val['ValueDisplayName']] = val[
-                            'ValueName']
-            defaultval = attr.get('DefaultValue', None)
-            defaultval = valtodisplay.get(attr['AttributeName'], {}).get(
-                defaultval, defaultval)
-            if attr['Type'] == 'Integer' and defaultval:
-                defaultval = int(defaultval)
-            if attr['Type'] == 'Boolean':
-                vals = [{'ValueDisplayName': 'True'},
-                        {'ValueDisplayName': 'False'}]
-            addon[attr['AttributeName']] = {
-                'default': defaultval,
-                'help': attr.get('HelpText', None),
-                'sortid': attr.get('DisplayOrder', None),
-                'possible': [x['ValueDisplayName'] for x in vals],
-            }
-        return addon, valtodisplay, displaytoval, reg
+        return self.oem.get_health(self, verbose)
 
     def get_bmc_configuration(self):
         """Get miscellaneous BMC configuration
@@ -988,6 +813,9 @@ class Command(object):
         # For now, this is a stub, no implementation for redfish currently
         return self.oem.set_bmc_configuration(changeset)
 
+    def set_system_configuration(self, changeset):
+        return self.oem.set_system_configuration(changeset, self)
+
     def clear_bmc_configuration(self):
         """Reset BMC to factory default
 
@@ -999,64 +827,7 @@ class Command(object):
             'Clear BMC configuration not supported in redfish yet')
 
     def get_system_configuration(self, hideadvanced=True):
-        return self._getsyscfg()[0]
-
-    def _getsyscfg(self):
-        biosinfo = self._do_web_request(self._biosurl, cache=False)
-        reginfo = ({}, {}, {}, {})
-        extrainfo = {}
-        valtodisplay = {}
-        self.attrdeps = {'Dependencies': [], 'Attributes': []}
-        if 'AttributeRegistry' in biosinfo:
-            overview = self._do_web_request('/redfish/v1/')
-            reglist = overview['Registries']['@odata.id']
-            reglist = self._do_web_request(reglist)
-            regurl = None
-            for cand in reglist.get('Members', []):
-                cand = cand.get('@odata.id', '')
-                candname = cand.split('/')[-1]
-                if candname == '':  # implementation uses trailing slash
-                    candname = cand.split('/')[-2]
-                if candname == biosinfo['AttributeRegistry']:
-                    regurl = cand
-                    break
-            if not regurl:
-                # Workaround a vendor bug where they link to a
-                # non-existant name
-                for cand in reglist.get('Members', []):
-                    cand = cand.get('@odata.id', '')
-                    candname = cand.split('/')[-1]
-                    candname = candname.split('.')[0]
-                    if candname == biosinfo[
-                            'AttributeRegistry'].split('.')[0]:
-                        regurl = cand
-                        break
-            if regurl:
-                reginfo = self._do_web_request(regurl)
-                for reg in reginfo.get('Location', []):
-                    if reg.get('Language', 'en').startswith('en'):
-                        reguri = reg['Uri']
-                        reginfo = self._get_biosreg(reguri)
-                        extrainfo, valtodisplay, _, self.attrdeps = reginfo
-        currsettings = {}
-        try:
-            pendingsettings = self._do_web_request(self._setbiosurl)
-        except exc.UnsupportedFunctionality:
-            pendingsettings = {}
-        pendingsettings = pendingsettings.get('Attributes', {})
-        for setting in biosinfo.get('Attributes', {}):
-            val = biosinfo['Attributes'][setting]
-            currval = val
-            if setting in pendingsettings:
-                val = pendingsettings[setting]
-            val = valtodisplay.get(setting, {}).get(val, val)
-            currval = valtodisplay.get(setting, {}).get(currval, currval)
-            val = {'value': val}
-            if currval != val['value']:
-                val['active'] = currval
-            val.update(**extrainfo.get(setting, {}))
-            currsettings[setting] = val
-        return currsettings, reginfo
+        return self.oem.get_system_configuration(hideadvanced, self)
 
     def clear_system_configuration(self):
         """Clear the BIOS/UEFI configuration
@@ -1085,80 +856,27 @@ class Command(object):
             parms = {'Action': 'Bios.ResetBios'}
         self._do_web_request(rb, parms)
 
-    def set_system_configuration(self, changeset):
-        while True:
-            try:
-                self._set_system_configuration(changeset)
-                return
-            except exc.RedfishError as re:
-                if 'etag' not in re.msgid.lower():
-                    raise
-
-    def _set_system_configuration(self, changeset):
-        currsettings, reginfo = self._getsyscfg()
-        rawsettings = self._do_web_request(self._biosurl, cache=False)
-        rawsettings = rawsettings.get('Attributes', {})
-        pendingsettings = self._do_web_request(self._setbiosurl)
-        etag = pendingsettings.get('@odata.etag', None)
-        pendingsettings = pendingsettings.get('Attributes', {})
-        dephandler = AttrDependencyHandler(self.attrdeps, rawsettings,
-                                           pendingsettings)
-        for change in list(changeset):
-            if change not in currsettings:
-                found = False
-                for attr in currsettings:
-                    if fnmatch(attr.lower(), change.lower()):
-                        found = True
-                        changeset[attr] = changeset[change]
-                    if fnmatch(attr.lower(),
-                               change.replace('.', '_').lower()):
-                        found = True
-                        changeset[attr] = changeset[change]
-                if found:
-                    del changeset[change]
-        for change in changeset:
-            changeval = changeset[change]
-            overrides, blameattrs = dephandler.get_overrides(change)
-            meta = {}
-            for attr in self.attrdeps['Attributes']:
-                if attr['AttributeName'] == change:
-                    meta = dict(attr)
-                    break
-            meta.update(**overrides.get(change, {}))
-            if meta.get('ReadOnly', False) or meta.get('GrayOut', False):
-                errstr = '{0} is read only'.format(change)
-                if blameattrs:
-                    errstr += (' due to one of the following settings: '
-                               '{0}'.format(','.join(sorted(blameattrs)))
-                               )
-                raise exc.InvalidParameterValue(errstr)
-            if (currsettings.get(change, {}).get('possible', [])
-                    and changeval not in currsettings[change]['possible']):
-                normval = changeval.lower()
-                normval = re.sub(r'\s+', ' ', normval)
-                if not normval.endswith('*'):
-                    normval += '*'
-                for cand in currsettings[change]['possible']:
-                    if fnmatch(cand.lower(), normval):
-                        changeset[change] = cand
-                        break
+    def set_net6_configuration(self, static_addresses=None, static_gateway=None, name=None):
+        patch = {}
+        if static_addresses is not None:
+            sa = []
+            patch['IPv6StaticAddresses'] = sa
+            for addr in static_addresses:
+                if '/' in addr:
+                    addr, plen = addr.split('/', 1)
                 else:
-                    raise exc.InvalidParameterValue(
-                        '{0} is not a valid value for {1} ({2})'.format(
-                            changeval, change, ','.join(
-                                currsettings[change]['possible'])))
-            if changeset[change] in reginfo[2].get(change, {}):
-                changeset[change] = reginfo[2][change][changeset[change]]
-            for regentry in reginfo[3].get('Attributes', []):
-                if change in (regentry.get('AttributeName', ''),
-                              regentry.get('DisplayName', '')):
-                    if regentry.get('Type', None) == 'Integer':
-                        changeset[change] = int(changeset[change])
-                    if regentry.get('Type', None) == 'Boolean':
-                        changeset[change] = _to_boolean(changeset[change])
-        redfishsettings = {'Attributes': changeset}
-        self._do_web_request(self._setbiosurl, redfishsettings, 'PATCH',
-                             etag=etag)
+                    plen = '64'
+                sa.append({
+                    'PrefixLength': int(plen),
+                    'Address': addr
+                })
+        if static_gateway:
+            patch['IPv6StaticDefaultGateways'] = [{
+                'Address': static_gateway,
+            }]
+        if patch:
+            nicurl = self._get_bmc_nic_url(name)
+            self._do_web_request(nicurl, patch, 'PATCH')
 
     def set_net_configuration(self, ipv4_address=None, ipv4_configuration=None,
                               ipv4_gateway=None, name=None):
@@ -1166,6 +884,9 @@ class Command(object):
         ipinfo = {}
         dodhcp = None
         netmask = None
+        if (ipv4_address is None and ipv4_configuration is None
+                and ipv4_gateway is None):
+            return
         if ipv4_address:
             if '/' in ipv4_address:
                 ipv4_address, cidr = ipv4_address.split('/')
@@ -1179,7 +900,7 @@ class Command(object):
             patch['IPv4StaticAddresses'] = [ipinfo]
             ipinfo['Gateway'] = ipv4_gateway
             ipv4_configuration = 'static'
-        if ipv4_configuration.lower() == 'dhcp':
+        if ipv4_configuration and ipv4_configuration.lower() == 'dhcp':
             dodhcp = True
             patch['DHCPv4'] = {'DHCPEnabled': True}
         elif (ipv4_configuration == 'static'
@@ -1198,6 +919,21 @@ class Command(object):
                     ipinfo['AddressOrigin'] = 'Static'
                 self._do_web_request(nicurl, patch, 'PATCH')
 
+    def get_net6_configuration(self, name=None):
+        nicurl = self._get_bmc_nic_url(name)
+        netcfg = self._do_web_request(nicurl, cache=False)
+        retdata = {}
+        saddrs = netcfg.get('IPv6StaticAddresses', ())
+        retdata['static_addrs'] = []
+        for sa in saddrs:
+            ca = '{}/{}'.format(sa['Address'], sa['PrefixLength'])
+            retdata['static_addrs'].append(ca)
+        gws = netcfg.get('IPv6StaticDefaultGateways', None)
+        if gws:
+            for gw in gws:
+                retdata['static_gateway'] = gw['Address']
+        return retdata
+
     def get_net_configuration(self, name=None):
         nicurl = self._get_bmc_nic_url(name)
         netcfg = self._do_web_request(nicurl, cache=False)
@@ -1279,6 +1015,63 @@ class Command(object):
     def get_inventory(self, withids=False):
         return self.oem.get_inventory(withids)
 
+    def get_location_information(self):
+        locationinfo = {}
+        for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+            chassisurl = chassis['@odata.id']
+            data = self._do_web_request(chassisurl)
+            locdata = data.get('Location', {})
+            postaladdress = locdata.get('PostalAddress', {})
+            placement = locdata.get('Placement', {})
+            contactinfo = locdata.get('Contacts', [])
+            currval = postaladdress.get('Room', '')
+            if currval:
+                locationinfo['room'] = currval
+            currval = postaladdress.get('Location', '')
+            if currval:
+                locationinfo['location'] = currval
+            currval = postaladdress.get('Building', '')
+            if currval:
+                locationinfo['building'] = currval
+            currval = placement.get('Rack', '')
+            if currval:
+                locationinfo['rack'] = currval
+            for contact in contactinfo:
+                contact = contact.get('ContactName', '')
+                if not contact:
+                    continue
+                if 'contactnames' not in locationinfo:
+                    locationinfo['contactnames'] = [contact]
+                else:
+                    locationinfo['contactnames'].append(contact)
+        return locationinfo
+
+    def set_location_information(self, room=None, contactnames=None,
+                                 location=None, building=None, rack=None):
+        locationinfo = {}
+        postaladdress = {}
+        placement = {}
+        if contactnames is not None:
+            locationinfo['Contacts'] = [
+                {'ContactName': x} for x in contactnames]
+        if room is not None:
+            postaladdress['Room'] = room
+        if location is not None:
+            postaladdress['Location'] = location
+        if building is not None:
+            postaladdress['Building'] = building
+        if rack is not None:
+            placement['Rack'] = rack
+        if postaladdress:
+            locationinfo['PostalAddress'] = postaladdress
+        if placement:
+            locationinfo['Placement'] = placement
+        if locationinfo:
+            for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+                chassisurl = chassis['@odata.id']
+                self._do_web_request(chassisurl, {'Location': locationinfo},
+                                     method='PATCH')
+
     @property
     def oem(self):
         if not self._oem:
@@ -1338,6 +1131,10 @@ class Command(object):
                         entries = self._do_web_request(entriesurl, cache=False)
                         newloginfo = self._do_web_request(lurl, cache=False)
             for log in entries.get('Members', []):
+                if ('Created' not in log and 'Message' not in log
+                        and 'Severity' not in log):
+                    # without any data, this log entry isn't actionable
+                    continue
                 record = {}
                 record['log_id'] = logid
                 parsedtime = parse_time(log.get('Created', ''))
@@ -1349,9 +1146,59 @@ class Command(object):
                     record['timestamp'] = log.get('Created', '')
                 record['message'] = log.get('Message', None)
                 record['severity'] = _healthmap.get(
-                    entries.get('Severity', 'Warning'), const.Health.Critical)
+                    log.get('Severity', 'Warning'), const.Health.Ok)
                 yield record
 
+    def _get_chassis_env(self, chassis):
+        chassisurl = chassis['@odata.id']
+        chassisinfo = self._do_web_request(chassisurl)
+        envurl = chassisinfo.get('EnvironmentMetrics', {}).get('@odata.id', '')
+        envmetric = self._do_web_request(envurl, cache=1)
+        retval = {
+            'watts': envmetric.get('PowerWatts', {}).get('Reading', None),
+            'inlet': envmetric.get('TemperatureCelsius', {}).get('Reading', None)
+        }
+        return retval
+
+    def get_average_processor_temperature(self):
+        cputemps = []
+        for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+            thermals = self._get_thermals(chassis)
+            for temp in thermals:
+                if temp.get('PhysicalContext', '') != 'CPU':
+                    continue
+                if temp.get('ReadingCelsius', None) is None:
+                    continue
+                cputemps.append(temp['ReadingCelsius'])
+        if not cputemps:
+            return  SensorReading(
+            None, {'name': 'Average Processor Temperature'}, value=None, units='°C',
+                   unavailable=True)
+        avgtemp = sum(cputemps) / len(cputemps)
+        return SensorReading(
+            None, {'name': 'Average Processor Temperature'}, value=avgtemp, units='°C')
+
+    def get_system_power_watts(self):
+        totalwatts = 0
+        for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+            envinfo = self._get_chassis_env(chassis)
+            totalwatts += envinfo['watts']
+        return totalwatts
+
+    def get_inlet_temperature(self):
+        inlets = []
+        for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+            envinfo = self._get_chassis_env(chassis)
+            inlets.append(envinfo['inlet'])
+        if inlets:
+            val = sum(inlets) / len(inlets)
+            unavail = False
+        else:
+            val = None
+            unavail = True
+        return SensorReading(
+                        None, {'name': 'Inlet Temperature'}, value=val, units='°C',
+                        unavailable=unavail)
     def get_sensor_descriptions(self):
         for sensor in natural_sort(self._sensormap):
             yield self._sensormap[sensor]
@@ -1428,6 +1275,16 @@ class Command(object):
         """
         return self.oem.apply_storage_configuration(cfgspec)
 
+    def check_storage_configuration(self, cfgspec=None):
+        """Evaluate a configuration for validity
+
+        This will check if configuration is currently available and, if given,
+        whether the specified cfgspec can be applied.
+        :param cfgspec: A pyghmi.storage.ConfigSpec describing desired oonfig
+        :return:
+        """
+        return self._oem.check_storage_configuration(cfgspec)
+
     def attach_remote_media(self, url, username=None, password=None):
         """Attach remote media by url
 
@@ -1444,9 +1301,12 @@ class Command(object):
         # identify the correct resource ahead of time.
         # As such it's OEM specific until the standard
         # provides a better way.
-        bmcinfo = self._do_web_request(self._bmcurl)
-        vmcoll = bmcinfo.get('VirtualMedia', {}).get('@odata.id', None)
-        vmurls = None
+        vmurls = []
+        vmcoll = self.sysinfo.get(
+            'VirtualMedia', {}).get('@odata.id', None)
+        if not vmcoll:
+            vmcoll = self.bmcinfo.get(
+                'VirtualMedia', {}).get('@odata.id', None)
         if vmcoll:
             vmlist = self._do_web_request(vmcoll)
             vmurls = [x['@odata.id'] for x in vmlist.get('Members', [])]
@@ -1456,29 +1316,57 @@ class Command(object):
             return
         for vmurl in vmurls:
             vminfo = self._do_web_request(vmurl, cache=False)
-            if vminfo['ConnectedVia'] != 'NotConnected':
+            if vminfo.get('ConnectedVia', None) != 'NotConnected':
                 continue
-            self._do_web_request(vmurl, {'Image': url, 'Inserted': True},
-                                 'PATCH')
+            inserturl = vminfo.get(
+                'Actions', {}).get(
+                    '#VirtualMedia.InsertMedia', {}).get('target', None)
+            if inserturl:
+                self._do_web_request(inserturl, {'Image': url})
+            else:
+                try:
+                    self._do_web_request(vmurl,
+                                         {'Image': url, 'Inserted': True},
+                                         'PATCH')
+                except exc.RedfishError as re:
+                    if re.msgid.endswith(u'PropertyUnknown'):
+                        self._do_web_request(vmurl, {'Image': url}, 'PATCH')
+                    else:
+                        raise
             break
 
     def detach_remote_media(self):
-        bmcinfo = self._do_web_request(self._bmcurl)
-        vmcoll = bmcinfo.get('VirtualMedia', {}).get('@odata.id', None)
         try:
             self.oem.detach_remote_media()
         except exc.BypassGenericBehavior:
             return
+        vmcoll = self.sysinfo.get('VirtualMedia', {}).get('@odata.id', None)
+        if not vmcoll:
+            bmcinfo = self._do_web_request(self._bmcurl)
+            vmcoll = bmcinfo.get('VirtualMedia', {}).get('@odata.id', None)
         if vmcoll:
             vmlist = self._do_web_request(vmcoll)
             vmurls = [x['@odata.id'] for x in vmlist.get('Members', [])]
             for vminfo in self._do_bulk_requests(vmurls):
                 vminfo, currl = vminfo
                 if vminfo['Image']:
-                    self._do_web_request(currl,
-                                         {'Image': None,
-                                          'Inserted': False},
-                                         method='PATCH')
+                    ejurl = vminfo.get(
+                        'Actions', {}).get(
+                            '#VirtualMedia.EjectMedia', {}).get('target', None)
+                    if ejurl:
+                        self._do_web_request(ejurl, {})
+                    else:
+                        try:
+                            self._do_web_request(currl,
+                                                 {'Image': None,
+                                                  'Inserted': False},
+                                                 method='PATCH')
+                        except exc.RedfishError as re:
+                            if re.msgid.endswith(u'PropertyUnknown'):
+                                self._do_web_request(currl, {'Image': None},
+                                                     method='PATCH')
+                            else:
+                                raise
 
     def upload_media(self, filename, progress=None, data=None):
         """Upload a file to be hosted on the target BMC
@@ -1503,6 +1391,8 @@ class Command(object):
                            update process.  Provide if
          :param bank: Indicate a target 'bank' of firmware if supported
         """
+        if progress is None:
+            progress = lambda x: True
         return self.oem.update_firmware(file, data, progress, bank)
 
     def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
