@@ -52,7 +52,7 @@ class EntryField(object):
 
 
 # General parameter parsing functions
-def parse_inventory_category(name, info, countable=True):
+def parse_inventory_category(name, info, asrock=False, countable=True):
     """Parses every entry in an inventory category
 
     For example: CPU, memory, PCI, drives
@@ -61,14 +61,18 @@ def parse_inventory_category(name, info, countable=True):
 
     :param name: the name of the parameter (e.g.: "cpu")
     :param info: a list of integers with raw data read from an IPMI requests
+    :param asrock: a boolean represents if RS160 with asrockrack or not
     :param countable: whether the data have an entries count field
     :returns: dict -- a list of entries in the category.
     """
     raw = info["data"][1:]
 
+    if name == "cpu" and asrock:
+        raw = info["data"]
+
     cur = 0
     if countable:
-        count = struct.unpack("B", raw[cur])[0]
+        count = bytearray(raw)[cur]
         cur += 1
     else:
         count = 0
@@ -76,17 +80,17 @@ def parse_inventory_category(name, info, countable=True):
 
     entries = []
     while cur < len(raw):
-        read, cpu = categories[name]["parser"](raw[cur:])
+        read, parser = categories[name]["parser"](raw[cur:])
         cur = cur + read
         # Account for discarded entries (because they are not present)
-        if cpu is None:
+        if parser is None:
             discarded += 1
             continue
         if not countable:
             # count by myself
             count += 1
-            cpu["index"] = count
-        entries.append(cpu)
+            parser["index"] = count
+        entries.append(parser)
 
     # TODO(avidal): raise specific exception to point that there's data left in
     # the buffer
@@ -94,7 +98,7 @@ def parse_inventory_category(name, info, countable=True):
         raise Exception
     # TODO(avidal): raise specific exception to point that the number of
     # entries is different than the expected
-    if count - discarded != len(entries):
+    if count - discarded != len(entries) and not asrock:
         raise Exception
     return entries
 
@@ -119,6 +123,9 @@ def parse_inventory_category_entry(raw, fields):
         value = struct.unpack_from(field.fmt, r)[0]
         read = struct.calcsize(field.fmt)
         bytes_read += read
+        if bytes_read > len(raw):
+            break
+
         r = r[read:]
         # If this entry is not actually present, just parse and then discard it
         if field.presence and not bool(value):
@@ -127,7 +134,7 @@ def parse_inventory_category_entry(raw, fields):
             continue
 
         if (field.fmt[-1] == "s"):
-            value = value.rstrip("\x00")
+            value = value.rstrip(b'\x00\xff')
         if (field.mapper and value in field.mapper):
             value = field.mapper[value]
         if (field.valuefunc):
@@ -142,3 +149,33 @@ def parse_inventory_category_entry(raw, fields):
     if discard:
         obj = None
     return bytes_read, obj
+
+
+def parse_bios_number_entry(raw):
+    """Parses the Bios number given a raw data.
+
+    :param raw: the raw data to the entry.
+
+    :returns: dict -- structure with read current and newest versions
+    """
+    new_major_version = struct.unpack_from("1B", raw, 25)[0]
+    new_minor_version = struct.unpack_from("1B", raw, 26)[0]
+    new_aux = struct.unpack_from("I", raw, 27)[0]
+
+    cur_major_version = struct.unpack_from("1B", raw, 31)[0]
+    cur_minor_version = struct.unpack_from("1B", raw, 32)[0]
+    cur_aux = struct.unpack_from("I", raw, 33)[0]
+
+    new_image_version = "%s.%s.%s" % (
+        str(new_major_version),
+        str(new_minor_version),
+        str(new_aux))
+    cur_image_version = "%s.%s.%s" % (
+        str(cur_major_version),
+        str(cur_minor_version),
+        str(cur_aux))
+
+    return {
+        'new_img_version': new_image_version,
+        'cur_img_version': cur_image_version
+    }
